#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#ifndef NULL
#define NULL 0x00
#endif 

#define APP_NAME        "gsa"
#define APP_VERSION     "0.1.8"
#define APP_YEAR        "2008"
#define APP_AUTHORS     "Jan Skalicky"
#define APP_COMMENTS    "GA framework for Genome Sequence analysis" 

/* switch definitions */
#define OPT_HELP1       "--help"
#define OPT_HELP        "-?"
#define OPT_VERBOSE     "-v"
#define OPT_POPSIZE     "-p="
#define OPT_ITER        "-i="
#define OPT_STOPCOND    "-q="
#define OPT_SCALING     "-r="
#define OPT_TRANSEXP    "-f="
#define OPT_MINIMIZE    "-m"
#define OPT_ALG_SELECT  "-as="
#define OPT_ALG_XOVER   "-ax="
#define OPT_ALG_MUTATE  "-am="
#define OPT_ALG_REPLACE "-ar="
#define OPT_P_XOVER     "-px="
#define OPT_P_MUTATE    "-pm="
#define OPT_TOURNSIZE   "-t="
#define OPT_STEADYPLACE "-s="
#define OPT_MILAMBDA    "-l="
#define OPT_ELITISM     "-e="
#define OPT_LINSCALL    "-z1="
#define OPT_LINSCALH    "-z2="


char buf[1024];

int opt_verbose    = 0;       /* verbose mode enabled/disabled */
int opt_popsize    = 100;     /* size of population */
int opt_iter       = 100;     /* number of iterations (generations/epochs) */
float opt_stopcond = 0.0;     /* stop condtition on fitness value reached (zero when not used) */
int opt_scaling    = 0;       /* scaling of fitness function */
float opt_transexp = 1.0;     /* transformation of (prescaled) fitness function by powering */
int opt_minimize   = 0;       /* we are solving maximization (0) or minimization (1) problem */

int opt_alg_select  = 0;      /* algorithm of used selection */
int opt_alg_xover   = 0;      /* algorithm of used crossover */
int opt_alg_mutate  = 0;      /* algorithm of used mutation */
int opt_alg_replace = 0;      /* algorithm of used population replacement strategy */

int opt_p_xover = 90;         /* percentual probability of crossover */
int opt_p_mutate = 10;        /* inversed probability of mutation */
float opt_tournsize = 4.0;    /* size of tournament in tournament selection */
int opt_steadyplace = 10;     /* percentual amount of population to be replaced in steady-state */
int opt_milambda = 100;       /* percentual ratio lambda/mi in mi-lambda replacement */
int opt_elitism = 3;          /* number of best individuals to keep in next population */
float opt_linscall = 1.0;     /* lower bound for linear scaling transformation */
float opt_linscalh = 100.0;   /* upper bound for linear scaling transformation */


#define FRAGMENT_MAXNUM 1024
#define FRAGMENT_MAXLEN 64
/* A set of fragmentary gene segments */
static char fragments[FRAGMENT_MAXNUM][FRAGMENT_MAXLEN];
static int fragment_count;    /* number of fragments used */
static int fragment_seqlen;   /* length of sequence if no optimalization done (sum of fragments length) */

struct tPoint
 {
	int x;
	int y;
 };

#define GENTYPE short   /* type of data in one gen - type of allele */

struct tIndividual   /* structure of individual */
 {
	float fitness;
	void *phenotype;
	GENTYPE *genotype;
 };

struct tPopulation   /* structure of population */
 {
	struct tIndividual *ind;
	int gencnt;     /* length of individuals' genotype */
	int indcnt;     /* number of individuals */
	int indfitmax;  /* index of best individual */
	int indfitmin;  /* index of worst individual */
	float fitsum;   /* sum of all individuals' fintess (precalculated for selection) */
 };

static struct tPopulation *population;   /* one generation population */

static GENTYPE *opmem1;   /* general-purpose memory for operators evaluations (eg. gen translate tables) */
static GENTYPE *opmem2;   /* these buffers should be allocated to the length of genotype before using operators */
static void *fitmem;      /* memory used by fitness calculation, allocated by fitness_init(), fitness_free() */
static int *gamem1;       /* general-purpose memory for genetic algorithm (eg. stochastic sampling) */
static int *gamem2;       /* these buffers should be allocated to the population size */

void help()
 {
	printf( "%s %s  Copyleft (c) %s %s\n", APP_NAME, APP_VERSION, APP_AUTHORS, APP_YEAR );
	printf( "================\n" );
	printf( "effect:   %s\n", APP_COMMENTS );
	printf( "usage:    %s ["OPT_VERBOSE"] ["OPT_POPSIZE"%%] ["OPT_ITER"%%] ["OPT_STOPCOND"#]" 
		    " ["OPT_SCALING"%%] ["OPT_TRANSEXP"#] ["OPT_MINIMIZE"]" "\n" 
		    " ["OPT_ALG_SELECT"%%] ["OPT_ALG_XOVER"%%] ["OPT_ALG_MUTATE"%%] ["OPT_ALG_REPLACE"%%]" 
			" ["OPT_P_XOVER"%%] ["OPT_P_MUTATE"%%]" " ["OPT_TOURNSIZE"#]" " ["OPT_STEADYPLACE"%%]" 
			" ["OPT_MILAMBDA"%%] ["OPT_ELITISM"%%]" " ["OPT_LINSCALL"#]" " ["OPT_LINSCALH"#]\n", APP_NAME );
	printf( "----------------\n" );
	printf( "<options>\n" );
	printf( " %s:      help screen\n", OPT_HELP );
	printf( " %s:      verbose mode\n", OPT_VERBOSE );
	printf( " %s%%:    size of population (100 (*))\n", OPT_POPSIZE );
	printf( " %s%%:    number of generations (100 (*), 0 for infinite)\n", OPT_ITER );
	printf( " %s#:    stop condition on this or better fitness value (0 if unused (*))\n", OPT_STOPCOND );
	printf( " %s%%:    scaling of fitness function (0=none (*), 1=linear scaling, 2=ranking)\n", OPT_SCALING );
	printf( " %s#:    transformation of (prescaled) fitness function by powering to (1 (*))\n", OPT_TRANSEXP );
	printf( " %s:      minimization problem (roulette/stochastic probabilities ~ 1/fitness)\n", OPT_MINIMIZE );
	printf( " %s%%:   alg. of selection (0=tournament (*), 1=roulette, 2=stochastic sample)\n", OPT_ALG_SELECT );
	printf( " %s%%:   alg. of crossover (0=rnd (*), 1=OX, 2=PMX, 3=cycle, 4=edge recombine)\n", OPT_ALG_XOVER );
	printf( " %s%%:   alg. of mutation (0=rnd (*), 1=insert, 2=swap, 3=inverse, 4=scramble)\n", OPT_ALG_MUTATE );
	printf( " %s%%:   population replacement (0=en-block (*), 1=steady-state, 2=mi-lamda)\n", OPT_ALG_REPLACE );
	printf( " %s%%:   probability of crossover, percentual (90 (*))\n", OPT_P_XOVER );
	printf( " %s%%:   probability of mutation, inversed (10 (*), 0 for none)\n", OPT_P_MUTATE );
	printf( " %s#:    size of tournament in tournament selection, floating point (4.0 (*))\n", OPT_TOURNSIZE );
	printf( " %s%%:    percentual amount of steady-state population to be replaced (10 (*))\n", OPT_STEADYPLACE );
	printf( " %s%%:    percentual ratio lambda/mi in mi-lambda replacement (100 (*))\n", OPT_MILAMBDA );
	printf( " %s%%:    number of best individuals to keep in next population (3 (*))\n", OPT_ELITISM );
	printf( " %s#:    lower bound for linear scaling, use positive values (1.0 (*))\n", OPT_LINSCALL );
	printf( " %s#:    upper bound for linear scaling, use positive values (100.0 (*))\n", OPT_LINSCALH );
	printf( "<I/O>\n");
	printf( " input:   instance (fragment#n ([ACGT]*) <LF> ...)\n" ); 
	printf( "----------------\n" );
	printf( "(*)       default option\n" ); 
	printf( "example:  %s %s %s500 %s2000 %s1 %s1 %s80 %s5 </home/me/fragments\n", APP_NAME, 
		    OPT_MINIMIZE, OPT_POPSIZE, OPT_ITER, OPT_ALG_MUTATE, OPT_ALG_REPLACE, OPT_P_XOVER, OPT_P_MUTATE);
 }

int parser( int argc, char* argv[] )   /* parse CMDLine and load default options */
 {
	int i, aux;
	for( i = 1; i < argc; ++i )
	 {
		if( !( strcmp ( argv[i], OPT_HELP1 ) ) )   {help(); return 1;}
		if( !( strcmp ( argv[i], OPT_HELP  ) ) )   {help(); return 1;}
		if( !( strcmp ( argv[i], OPT_VERBOSE ) ) ) {opt_verbose = 1; continue;}
		if( !( strcmp ( argv[i], OPT_MINIMIZE ) ) ) {opt_minimize = 1; continue;}

		/* options without value */

		/* options with value */
		aux = strlen( OPT_POPSIZE );
		if( !( strncmp( argv[i], OPT_POPSIZE, aux ) ) ) {opt_popsize = atoi( argv[i] + aux ); continue;}
		aux = strlen( OPT_ITER );
		if( !( strncmp( argv[i], OPT_ITER, aux ) ) ) {opt_iter = atoi( argv[i] + aux ); continue;}
		aux = strlen( OPT_STOPCOND );
		if( !( strncmp( argv[i], OPT_STOPCOND, aux ) ) ) {opt_stopcond = atof( argv[i] + aux ); continue;}
		aux = strlen( OPT_SCALING );
		if( !( strncmp( argv[i], OPT_SCALING, aux ) ) ) {opt_scaling = atoi( argv[i] + aux ); continue;}
		aux = strlen( OPT_TRANSEXP );
		if( !( strncmp( argv[i], OPT_TRANSEXP, aux ) ) ) {opt_transexp = atof( argv[i] + aux ); continue;}
		aux = strlen( OPT_ALG_SELECT );
		if( !( strncmp( argv[i], OPT_ALG_SELECT, aux ) ) ) {opt_alg_select = atoi( argv[i] + aux ); continue;}
		aux = strlen( OPT_ALG_XOVER );
		if( !( strncmp( argv[i], OPT_ALG_XOVER, aux ) ) ) {opt_alg_xover = atoi( argv[i] + aux ); continue;}
		aux = strlen( OPT_ALG_MUTATE );
		if( !( strncmp( argv[i], OPT_ALG_MUTATE, aux ) ) ) {opt_alg_mutate = atoi( argv[i] + aux ); continue;}
		aux = strlen( OPT_ALG_REPLACE );
		if( !( strncmp( argv[i], OPT_ALG_REPLACE, aux ) ) ) {opt_alg_replace = atoi( argv[i] + aux ); continue;}
		aux = strlen( OPT_P_MUTATE );
		if( !( strncmp( argv[i], OPT_P_MUTATE, aux ) ) ) {opt_p_mutate = atoi( argv[i] + aux ); continue;}
		aux = strlen( OPT_P_XOVER );
		if( !( strncmp( argv[i], OPT_P_XOVER, aux ) ) ) {opt_p_xover = atoi( argv[i] + aux ); continue;}
		aux = strlen( OPT_TOURNSIZE );
		if( !( strncmp( argv[i], OPT_TOURNSIZE, aux ) ) ) {opt_tournsize = atof( argv[i] + aux ); continue;}
		aux = strlen( OPT_STEADYPLACE );
		if( !( strncmp( argv[i], OPT_STEADYPLACE, aux ) ) ) {opt_steadyplace = atoi( argv[i] + aux ); continue;}
		aux = strlen( OPT_MILAMBDA );
		if( !( strncmp( argv[i], OPT_MILAMBDA, aux ) ) ) {opt_milambda = atoi( argv[i] + aux ); continue;}
		aux = strlen( OPT_ELITISM );
		if( !( strncmp( argv[i], OPT_ELITISM, aux ) ) ) {opt_elitism = atoi( argv[i] + aux ); continue;}
		aux = strlen( OPT_LINSCALL );
		if( !( strncmp( argv[i], OPT_LINSCALL, aux ) ) ) {opt_linscall = atof( argv[i] + aux ); continue;}
		aux = strlen( OPT_LINSCALH );
		if( !( strncmp( argv[i], OPT_LINSCALH, aux ) ) ) {opt_linscalh = atof( argv[i] + aux ); continue;}

		/* default options (without minus prefix): */
	 }
	return 0;
}

void memerror()
 {
	fputs( "error: not enough memory to allocate buffer\n", stderr );
	exit( EXIT_FAILURE );   /* terminate program if out of memory */
 }

#define GEN(x, y, z) (x->ind[y].genotype[z])  /* gen z of individual y in population x */

/* allocate memory for population with popsize individuals, each with gencnt gens in chromosome */
struct tPopulation *popinit( int popsize, int gencnt )
 {
	int i;
	struct tPopulation *res;

	res = malloc( sizeof(*res) );
	if( res == NULL ) memerror();

	res->indcnt = popsize;
	res->gencnt = gencnt;
	res->ind = calloc( res->indcnt, sizeof(*res->ind) );
	if( res->ind == NULL ) memerror();
	for( i = 0; i < res->indcnt; ++i )
	 {
		res->ind[i].genotype = calloc( res->gencnt, sizeof(*res->ind[i].genotype) );
		if( res->ind[i].genotype == NULL ) memerror();
		res->ind[i].phenotype = NULL;   /* no phenotype representation present */
	 } 
	return res;
 }

/* deallocate memory for population */
void popkill( struct tPopulation *pop )
 {
	int i; 

	if( pop->ind != NULL )
	 {
		for( i = 0; i < pop->indcnt; ++i )
		 {
			if( pop->ind[i].genotype  != NULL ) free( pop->ind[i].genotype  );
			if( pop->ind[i].phenotype != NULL ) free( pop->ind[i].phenotype );
		 }
		free( pop->ind );
	 }
	free( pop );
 }

int cmp_tPoint_x( void const *a, void const *b )
 {
	return (*(struct tPoint *)a).x == (*(struct tPoint *)b).x ? 0 :
		   (*(struct tPoint *)a).x >  (*(struct tPoint *)b).x ? 1 : -1;
 }

/* fill population genotypes with random permutations */
void popfill_randperm( struct tPopulation *pop )
 {
	int i, j, aux;
	struct tPoint *rndkeys;

	rndkeys = calloc( pop->gencnt, sizeof(*rndkeys) );
	if( rndkeys == NULL ) memerror();
	
	for( i = 0; i < pop->indcnt; ++i )
	 {
		aux = 0;
		for( j = 0; j < pop->gencnt; ++j )
	 	 {
			rndkeys[j].x = rand();   /* random keys for inversions */
			rndkeys[j].y = aux++;    /* permutation w/o inversions */
		 }
		/* randomize permutation according to keys */
		qsort( (void *)rndkeys, pop->gencnt, sizeof(*rndkeys), &cmp_tPoint_x );
		/* store it as genotype */
		for( j = 0; j < pop->gencnt; ++j ) GEN( pop, i, j ) = rndkeys[j].y;
	 }
	free( rndkeys );
 }

/* calculate individual's fitness and optionally validate value and phenotype */
/* USE POSITIVE VALUES */
float fitness( struct tIndividual *ind, char validate )
 {
	int i, j, k;
	int pos, aux, bux;
	char *seq;   /* constructed Genome Sequence */
	float res;
	
	seq = (char *)fitmem;
	pos = 0;   /* first position in seq to be used */
	for( i = 0; i < fragment_count; ++i )
	 {   /* for each fragment */
		aux = ind->genotype[i];
		bux = strlen( fragments[aux] );
		for( j = bux; j > 0; --j )
		 {   /* for each possible common part (seq.suffix = fragment.prefix), longest first */
			if( pos - j < 0 ) continue;
			if( strncmp( seq + pos - j, fragments[aux], j ) == 0 ) break;
		 }
		for( k = j; k < bux; ++k )
		 {   /* paste fragment suffix onto seq */
			seq[pos] = fragments[aux][k];
			pos++;
		 }
	 }
	seq[pos] = '\0';

	/* fitness value from pos (length of final sequence) */
	res = pos;

	/* optional fitness set and phenotype validation */
	if( validate )
	 {
		ind->fitness = res;
		if( ind->phenotype != NULL ) free( ind->phenotype );
		ind->phenotype = malloc( pos+1 );
		if( ind->phenotype == NULL ) memerror();
		strcpy( ind->phenotype, seq );
	 }
	return res;
 }

void fitness_init()
 {
	fitmem = malloc( fragment_seqlen + 1 );
	if( fitmem == NULL ) memerror();
 }

void fitness_free()
 {
	free( fitmem );
 }

int cmp_fitness_max( void const *a, void const *b )
 {
	return (*(struct tIndividual *)a).fitness == (*(struct tIndividual *)b).fitness ? 0 :
		   (*(struct tIndividual *)a).fitness <  (*(struct tIndividual *)b).fitness ? 1 : -1;
 }

int cmp_fitness_min( void const *a, void const *b )
 {
	return (*(struct tIndividual *)a).fitness == (*(struct tIndividual *)b).fitness ? 0 :
		   (*(struct tIndividual *)a).fitness >  (*(struct tIndividual *)b).fitness ? 1 : -1;
 }

/* print genotype */
void genprint( FILE *out, struct tIndividual const *ind, int gencnt )
 {
	int i;

	for( i = 0;; )
	 {
		fprintf( out, "%d", ind->genotype[i] );
		if( ++i >= gencnt ) break;
		fputc( ' ', out );
	 }
 }

/* print phenotype */
void pheprint( FILE *out, struct tIndividual const *ind, int gencnt )
 {
	if( ind->phenotype != NULL ) fputs( (char *)ind->phenotype, out );
	else fputs( "N/A", out );
 }

/* print individual */
void indprint( FILE *out, struct tIndividual const *ind, int gencnt )
 {
	fputs( "genes = ", out );
	genprint( out, ind, gencnt );
	fprintf( out, "; fitness = %5.2f", ind->fitness );
	fputs( "; phenotype = ", out );
	pheprint( out, ind, gencnt );
 }

/* print population */
void popprint( FILE *out, struct tPopulation const *pop )
 {
	int i;

	for( i = 0; i < pop->indcnt; ++i )
	 {
		fprintf( out, "ind %2d: ", i );
		indprint( out, &pop->ind[i], pop->gencnt );
		fputc( '\n', out );
	 }
 }

/* sort population by quality (descending order by default) */
void popsort( struct tPopulation *pop, int asc )
 {
	/* greater fitness means better quality, because minimization problem has fitness values inverted already */
	if( !asc ) qsort( pop->ind, pop->indcnt, sizeof(*pop->ind), &cmp_fitness_max );
	/* but in some cases, ascending order can be required */
	else       qsort( pop->ind, pop->indcnt, sizeof(*pop->ind), &cmp_fitness_min );
 }

/***********************************************************************/
/* functions select_... return one parent selected by chosen technique */

struct tIndividual *select_roulette( struct tPopulation const *pop )
 {
	int i;
	float rnd, sum = 0.0;

	rnd = (float)rand() / RAND_MAX * pop->fitsum;
	for( i = 0; i < pop->indcnt; ++i )
	 {
		sum += pop->ind[i].fitness;
		if( sum >= rnd ) return &pop->ind[i];
	 }
	return &pop->ind[pop->indcnt-1];   /* singularity on sum of floats */
 }

struct tIndividual *select_tournament( struct tPopulation const *pop )
 {
	int i;
	int aux;
	int sel;
	float tournsize;
	static float tourncredit = 0;   /* used if tournament size is floating point */

	tournsize = floor( opt_tournsize + tourncredit );   /* eg. tournament size 3.25 means 3x size 3 and 1x size 4 */
	tourncredit += opt_tournsize - tournsize;           /* this is what remains to next call */

	sel = rand() % pop->indcnt;
	for( i = 1; i < tournsize; ++i )
	 {
		aux = rand() % pop->indcnt;
		if( pop->ind[aux].fitness > pop->ind[sel].fitness ) sel = aux;
	 }
	return &pop->ind[sel];
 }

/****************************************************************************************************/
/* functions xover_... crossover parents and write non-NULL childs, return number of childs written */

int xover_OX( struct tIndividual const *par1, struct tIndividual const *par2, 
		      struct tIndividual *child1, struct tIndividual *child2, int gencnt )
 {
	int i, j;
	int cut1, cut2;
	int res = 0;

	/* cutting section between cut1 and cut2 */
	cut1 = rand() % (gencnt + 1);
	cut2 = rand() % (gencnt + 1);
	if( cut1 > cut2 ) cut1 ^= cut2, cut2 ^= cut1, cut1 ^= cut2;   /* swap to have cut1 <= cut2 */

	/* opmem will hold whether gen value is used or not yet */
	memset( opmem1, 0, gencnt * sizeof(*opmem1) );   /* reset */
	memset( opmem2, 0, gencnt * sizeof(*opmem2) );
	for( i = cut1; i < cut2; ++i )
	 {
		if( child1 != NULL )
		 {
			child1->genotype[i] = par1->genotype[i];
			opmem1[child1->genotype[i]] = 1;
		 }
		if( child2 != NULL )
		 {
			child2->genotype[i] = par2->genotype[i];
			opmem2[child2->genotype[i]] = 1;
		 }
 	 }

	/* complete genotypes from other parent */
	for( i = j = cut2;; )
	 {
		if( cut2 >= gencnt ) {cut2 = 0; if( cut2 == cut1 ) break;}
		if( child1 != NULL ) 
		 {
			do if( i >= gencnt ) i = 0; while( opmem1[par2->genotype[i++]] );   /* find not used gen */
			child1->genotype[cut2] = par2->genotype[i-1];                       /* and write it */
			/*opmem1[child1->genotype[cut2]] = 1;*/
		 }
		if( child2 != NULL )
		 {
			do if( j >= gencnt ) j = 0; while( opmem2[par1->genotype[j++]] );
			child2->genotype[cut2] = par1->genotype[j-1];
			/*opmem2[child2->genotype[cut2]] = 1;*/
		 }
		if( ++cut2 == cut1 ) break;
	 }

	if( child1 != NULL ) ++res;
	if( child2 != NULL ) ++res;
	return res;
 }

int xover_PMX( struct tIndividual const *par1, struct tIndividual const *par2, 
		       struct tIndividual *child1, struct tIndividual *child2, int gencnt )
 {
	int i, j;
	int cut1, cut2;
	int res = 0;

	/* cutting section between cut1 and cut2 */
	cut1 = rand() % (gencnt + 1);
	cut2 = rand() % (gencnt + 1);
	if( cut1 > cut2 ) cut1 ^= cut2, cut2 ^= cut1, cut1 ^= cut2;   /* swap to have cut1 <= cut2 */

	/* calculate translate tables */
	memset( opmem1, -1, gencnt * sizeof(*opmem1) );   /* reset translate tables */
	memset( opmem2, -1, gencnt * sizeof(*opmem2) );   /* opmem1 will translate par1 genes, opmem2 par2 genes */
	for( i = cut1; i < cut2; ++i )
	 {
		opmem1[par1->genotype[i]] = par2->genotype[i];   /* mapping between cutting sections */
		opmem2[par2->genotype[i]] = par1->genotype[i];
 	 }
	for( i = cut1; i < cut2; ++i )
	 {
		j = par1->genotype[i];
		if( opmem2[j] == -1 )   /* gen mapped only from par1 section, not backwards */
		 {
			while( opmem1[j] != -1 ) j = opmem1[j];   /* find gen value for mapping */
			opmem1[j] = par1->genotype[i];            /* set new mapping from par1 */
			opmem2[par1->genotype[i]] = j;            /* and inversed mapping from par2 */
		 }
	 }
	
	/* generate childs */
	if( child1 != NULL )
	 {
		for( i = 0; i < gencnt; ++i ) /* map using translate tables (if table cell unset, map directly) */
			child1->genotype[i] = opmem1[par1->genotype[i]] != -1 ? opmem1[par1->genotype[i]] : par1->genotype[i];
		++res;
	 }
	if( child2 != NULL )
	 {
		for( i = 0; i < gencnt; ++i )
			child2->genotype[i] = opmem2[par2->genotype[i]] != -1 ? opmem2[par2->genotype[i]] : par2->genotype[i];
		++res;
	 }
	return res;
 }

int xover_cycle( struct tIndividual const *par1, struct tIndividual const *par2, 
		         struct tIndividual *child1, struct tIndividual *child2, int gencnt )
 {
	/* TODO */
	return 0;
 }

int xover_ERX( struct tIndividual const *par1, struct tIndividual const *par2, 
		       struct tIndividual *child1, struct tIndividual *child2, int gencnt )
 {
	/* TODO */
	return 0;
 }

/**************************************************************/
/* functions mutate_... mutate individual by chosen technique */

void mutate_insert( struct tIndividual *ind, int gencnt )
 {
	int al1, al2;
	GENTYPE tmp;

	/* cutting section between cut1 and cut2 */
	al1 = rand() % gencnt;
	al2 = rand() % gencnt;
	if( al1 > al2 ) al1 ^= al2, al2 ^= al1, al1 ^= al2;   /* swap to have al1 <= al2 */
	if( al1 == al2 ) return;   /* nothing to insert */

	/* insert, shifting */
	tmp = ind->genotype[al2--];
	while( al1 < al2 )
	 {
		ind->genotype[al2+1] = ind->genotype[al2];
		--al2;
	 }
	ind->genotype[al1+1] = tmp;
 }

void mutate_swap( struct tIndividual *ind, int gencnt )
 {
	int al1, al2;
	GENTYPE tmp;

	al1 = rand() % gencnt;
	al2 = rand() % gencnt;
	tmp = ind->genotype[al2];
	ind->genotype[al2] = ind->genotype[al1];
	ind->genotype[al1] = tmp;
 }

void mutate_inversion( struct tIndividual *ind, int gencnt )
 {
	int al1, al2;
	GENTYPE tmp;

	/* cutting section between cut1 and cut2 */
	al1 = rand() % gencnt;
	al2 = rand() % gencnt;
	if( al1 > al2 ) al1 ^= al2, al2 ^= al1, al1 ^= al2;   /* swap to have al1 <= al2 */
	if( al1 == al2 ) return;   /* nothing to inverse */

	/* inverse substring */
	while( al1 < al2 )
	 {
		tmp = ind->genotype[al2];
		ind->genotype[al2] = ind->genotype[al1];
		ind->genotype[al1] = tmp;
		++al1;
		--al2;
	 }
 }

void mutate_scramble( struct tIndividual *ind, int gencnt )
 {
	int i, j, k, l;
	GENTYPE tmp;

	/* get random subset of genes */
	for( i = 0, j = 0; i < gencnt; ++i )
	 {
		opmem1[i] = rand() & 0x01;                        /* mark chosen ones */
		if( opmem1[i] ) opmem2[j++] = ind->genotype[i];   /* save this subset */
	 }

	/* shuffle them */
	for( i = 0; i < j; ++i )   /* repeat (number of extracted gens)x double pick - each gen has 86% to be swapped */
	 {
		k = rand() % j;   /* random pick one */
		l = rand() % j;   /* random pick two */
		tmp = opmem2[l];
		opmem2[l] = opmem2[k];
		opmem2[k] = tmp;  /* and swap them */
	 }

	/* return shuffled subset back to chromosome */
	for( i = 0, j = 0; i < gencnt; ++i )
	 {
		if( opmem1[i] ) ind->genotype[i] = opmem2[j++];
	 }
 }

void gencopy( struct tIndividual *to, struct tIndividual const *from, int start, int end )
 {
	memcpy( to->genotype + sizeof(*to->genotype) * start, 
		    from->genotype + sizeof(*from->genotype) * start, 
		    sizeof(*to->genotype) * (end - start + 1 ) );
 }

/* selection and recombination until new population has popnewsize new individuals */
/* popnew is expected to be empty at call */
void ga_reproduction( struct tPopulation const *popold, struct tPopulation *popnew, int popnewsize )
 {
	struct tIndividual *par[2], *child[2];
	int childs, newinds = 0;
	int i;
	int sampled = 0;   /* whether stochastic sampling has been done */

	while( newinds < popnewsize )
	 {
		/** selection (parents in par[0], par[1]) **/
		switch( opt_alg_select )
		 {
		  case 0:   /* tournament */
			do {
				par[0] = select_tournament( popold );
				par[1] = select_tournament( popold );
			 } while( par[0] == par[1] );
			 break;
		  case 1:   /* roulette wheel */
			do {
				par[0] = select_roulette( popold );
				par[1] = select_roulette( popold );
			 } while( par[0] == par[1] );
			 break;
		  case 2:   /* stochastic sampling */
			if( !sampled )   /* sample population to auxiliary memory once */
			 {
				float angle, phase = 0, sum = 0; 
				int cur = rand() % popold->indcnt;

				angle = popold->fitsum / popold->indcnt;
				for( i = 0; i < popnewsize; ++i )
				 {
					while( sum < phase )
					 {
						cur = (cur+1) % popold->indcnt;
						sum += popold->ind[cur].fitness;
					 }
					gamem1[i] = cur;   /* first parents are sampled properly according to fitness value */
					phase += angle;
				 }
				sampled = 1;
			 }
			/* second parents are picked randomly from sampled selection */
			par[0] = &popold->ind[gamem1[newinds]];
			par[1] = &popold->ind[gamem1[rand() % popnewsize]];
			break;
		  default:   /* no defined selection - take parent with index 0 */
			par[0] = par[1] = &popold->ind[0];
		 }
		
		/* referencing childs (array overflow check) */
		child[0] = &popnew->ind[newinds];
		if( newinds+1 < popnewsize ) child[1] = &popnew->ind[newinds+1]; else child[1] = NULL;

		/** crossover **/
		if( rand() % 100 >= opt_p_xover ) goto no_xover;   /* process probability of crossover */
		switch( opt_alg_xover == 0 ? rand() % 4 + 1 : opt_alg_xover )   /* 0 is random pick of technique */
		 {
		  case 1:   /* OX - order 1 rossover */ 
			childs = xover_OX( par[0], par[1], child[0], child[1], popold->gencnt ); break;
		  case 2:   /* PMX - partially mapped */
			childs = xover_PMX( par[0], par[1], child[0], child[1], popold->gencnt ); break;
		  case 3:   /* cycle crossover */
			childs = xover_cycle( par[0], par[1], child[0], child[1], popold->gencnt ); break;
		  case 4:   /* ERX - edge recombination */
			childs = xover_ERX( par[0], par[1], child[0], child[1], popold->gencnt ); break;
no_xover:
		  default:
			childs = 0;
			/* no defined crossover - just copy chromosomes */
			if( child[0] != NULL ) {gencopy( child[0], par[0], 0, popold->gencnt-1 ); ++childs;}
			if( child[1] != NULL ) {gencopy( child[1], par[1], 0, popold->gencnt-1 ); ++childs;}
		 }

		/** mutation (on each child apart) **/
		for( i = 0; i < childs; ++i )
		 {
			if( child[i] == NULL ) continue;
			if( opt_p_mutate == 0 ) continue;   /* mutation disabled */
			if( rand() % opt_p_mutate != 0 ) goto no_mutate;   /* process probability of mutation */
			switch( opt_alg_mutate == 0 ? rand() % 4 + 1 : opt_alg_mutate )   /* 0 is random pick of technique */
			 {
			  case 1: mutate_insert   ( child[i], popold->gencnt ); break;   /* insert mutation */
			  case 2: mutate_swap     ( child[i], popold->gencnt ); break;   /* swap */
			  case 3: mutate_inversion( child[i], popold->gencnt ); break;   /* inversion */
			  case 4: mutate_scramble ( child[i], popold->gencnt ); break;   /* scramble */
no_mutate:
			  default:   /* no defined mutation - do nothing */
				;
			 }
		 }

		/* some new individuals written */
		newinds += childs;
	 }
	popnew->indcnt = newinds;
 }

/* population control - replace/adjust pop with popnew individuals (popnew can be modified, eg. sorted) */
void ga_replace( struct tPopulation *pop, struct tPopulation *popnew )
 {
	int i;
	int aux;

	switch( opt_alg_replace )
	 {
	  case 0:   /* en-block */
		if( opt_elitism > 0 ) popsort( pop, 0 );   /* on elitism enabled, best individuals won't be replaced */
		for( i = opt_elitism; i < pop->indcnt; ++i )
		 {
			/* copy rest of population */
			gencopy( &pop->ind[i], &popnew->ind[i], 0, pop->gencnt-1 );
		 }
	  	break;
	  case 1:   /* steady-state */
		popsort( pop, 0 );                 /* old population has fitness values already transformed according to prolem type (min/max) */
		for( i = 0; i < popnew->indcnt; ++i ) popnew->ind[i].fitness = fitness( &popnew->ind[i], 0 );
		popsort( popnew, opt_minimize );   /* new population with fresh fitness values must be sorted in the same mean to merge arrays */
		aux = opt_popsize * opt_steadyplace / 100;   /* absolute number of individuals to be replaced */
		for( i = 0; i < aux; ++i )
		 {
			if( pop->indcnt - i - 1 < 0 || i >= popnew->indcnt ) break;
			/* replace the worst individuals from pop by the best from popnew */
			gencopy( &pop->ind[pop->indcnt - i - 1], &popnew->ind[i], 0, pop->gencnt-1 );
		 }
		break;
	  case 2:   /* mi-lambda */
		/* TODO */
		break;
	  default:   /* no defined replacement (keep old population) */
		;
	 }
 }

/* genetic algorithm (pop is working population) */
int ga( struct tPopulation *pop )
 {
	int i, j;
	float fitaux;
	int stop = 0;

	struct tPopulation *offsprings;   /* auxiliary pop for reproduction */

	/* misc init */
	fitness_init();
	offsprings = popinit( pop->indcnt, pop->gencnt );
	opmem1 = calloc( pop->gencnt, sizeof( *opmem1 ) );
	opmem2 = calloc( pop->gencnt, sizeof( *opmem2 ) );
	if( opmem1 == NULL || opmem2 == NULL ) memerror();
	gamem1 = calloc( pop->indcnt, sizeof( *gamem1 ) );
	gamem2 = calloc( pop->indcnt, sizeof( *gamem2 ) );
	if( gamem1 == NULL || gamem2 == NULL ) memerror();

	/* first pop */
	popfill_randperm( pop );

	/** main cycle **/
	for( i = 1;; ++i )
	 { /* count fitness */
		pop->indfitmax = 0;
		pop->indfitmin = 0;
		pop->fitsum = 0;
		for( j = 0; j < pop->indcnt; ++j )
		 {
#ifdef DEBUG
		 fitaux = fitness( &pop->ind[j], 1 );
#else
		 fitaux = fitness( &pop->ind[j], 0 );
#endif
			pop->ind[j].fitness = fitaux;
			pop->fitsum += fitaux;
			/* find best/worst individual */
			if( fitaux > pop->ind[pop->indfitmax].fitness ) pop->indfitmax = j;
			if( fitaux < pop->ind[pop->indfitmin].fitness ) pop->indfitmin = j;
		 }

#ifdef DEBUG
		/* print whole pop (sorted by fitness) */
		fprintf( stderr, "GENERATION %d:\n", i );
		popprint( stderr, pop );
#endif
		if( opt_verbose ) fprintf( stdout, "generation %d best: %.2f\n", i, 
			opt_minimize ? pop->ind[pop->indfitmin].fitness : pop->ind[pop->indfitmax].fitness );

		/* solution good enough? (stop after fitness inversion, because ind with maximal value is expected as best) */
		if( opt_stopcond != 0 )
		 {
			if(  opt_minimize ) {if( pop->ind[pop->indfitmin].fitness <= opt_stopcond ) stop = 1;}
			else                {if( pop->ind[pop->indfitmax].fitness >= opt_stopcond ) stop = 1;}
		 }

		/* fitness inversion (minimization problem) */
		if( opt_minimize )
		 {   /* in case of minimization problem, fitness values are inverted, so without another transformation,
			    probabilities of selection are weighted by inversed (minimization) or direct (maximization) values */
			pop->indfitmax ^= pop->indfitmin ^= pop->indfitmax ^= pop->indfitmin;   /* swap */
			pop->fitsum = 0;
			for( j = 0; j < pop->indcnt; ++j )
			 {
				fitaux = 1.0 / pop->ind[j].fitness;
				pop->ind[j].fitness = fitaux;
				pop->fitsum += fitaux;
			 }
		 }

		if( stop ) break;                             /* break on stop condition */
		if( i >= opt_iter && opt_iter != 0 ) break;   /* break on number of iterations */
		
		/* fitness scaling */
		switch( opt_scaling )
		 {
			float sclrange, fitrange, fitltemp;
		  case 1:   /* linear scaling */
			sclrange = opt_linscalh - opt_linscall;
			fitltemp = pop->ind[pop->indfitmin].fitness;
			fitrange = pop->ind[pop->indfitmax].fitness - fitltemp;
			if( fitrange == 0.0 ) fitrange = 1.0;
			pop->fitsum = 0;
			for( j = 0; j < pop->indcnt; ++j )
			 {
				fitaux = opt_linscall + (pop->ind[j].fitness - fitltemp) * sclrange / fitrange;
				pop->ind[j].fitness = fitaux;
				pop->fitsum += fitaux;
			 }
			break;
		  case 2:   /* ranking */
			popsort( pop, 0 );   /* sort by quality in descending order */
			for( j = 0; j < pop->indcnt; ++j )
			 {
				fitaux = pop->indcnt - j;   /* rank from 1 to indcnt */
				pop->ind[j].fitness = fitaux;
				pop->fitsum += fitaux;
			 }
			pop->indfitmax = 0;
			pop->indfitmin = pop->indcnt - 1;
			break;
		  default:   /* no defined scaling - do nothing */
			;
		 }
		/* fitness powering */
		if( opt_transexp != 1.0 )
		 {
			pop->fitsum = 0;
			for( j = 0; j < pop->indcnt; ++j )
			 {
				fitaux = pow( pop->ind[j].fitness, opt_transexp );
				pop->ind[j].fitness = fitaux;
				pop->fitsum += fitaux;
			 }
		 }

		/* pop reproduction */
		ga_reproduction( pop, offsprings, pop->indcnt );

		/* pop replacement */
		ga_replace( pop, offsprings );
	 }

	/* validate best individual's phenotype and fitness value */
	fitness( &population->ind[population->indfitmax], 1 );

	/* deallocations */
	fitness_free();
	popkill( offsprings );
	free( opmem1 );
	free( opmem2 );
	free( gamem1 );
	free( gamem2 );

	return i;
 }

int main( int argc, char *argv[] )
 {
	int i, j;
	int gcnt;

	if( parser( argc, argv ) )   /* parse CMDline; some commands force quit (return value 1) */
		return 0;

	/* random number generator initialization */
	srand( time( NULL ) );

	/* read fragments */
	fragment_seqlen = 0;
	for( i = 0; i < FRAGMENT_MAXNUM; ++i )
	 {
		if( fgets( buf, sizeof(buf), stdin ) == NULL ) break;
		for( j = 0; j < FRAGMENT_MAXLEN-1; ++j )
		 {
			if( buf[j] == '\n' || buf[j] == '\0' || buf[j] == '\r' ) break;
			fragments[i][j] = buf[j];
		 }
		fragments[i][j] = '\0';
		fragment_seqlen += strlen( fragments[i] );
	 }
	fragment_count = i;

	/* solve */
	if( fragment_count > 0 )
	 {
		/* allocate working population */
		population = popinit( opt_popsize, fragment_count );

		/* genetic algorithm */
		gcnt = ga( population );

		/* print solution */
		if( opt_verbose )
		 {
			fprintf( stdout, "best fitness found: %.2f\n", population->ind[population->indfitmax].fitness );
			fprintf( stdout, "phenotype: " );
			pheprint( stdout, &population->ind[population->indfitmax], population->gencnt ); 
			fputc( '\n', stdout );
			fprintf( stdout, "genotype: " );
			genprint( stdout, &population->ind[population->indfitmax], population->gencnt ); 
			fputc( '\n', stdout );
			fprintf( stdout, "generations required: %d\n", gcnt );
		 } else {
			fprintf( stdout, "%.2f ", population->ind[population->indfitmax].fitness );
			pheprint( stdout, &population->ind[population->indfitmax], population->gencnt );
			fputc( '\n', stdout );
		 }

		/* deallocate population */
		popkill( population );
	 }
	else if( opt_verbose ) fputs( "no fragments on input, quiting...\n", stderr ); 
	return 0;
 }

